/*
								+----------------------------------+
								|                                  |
								|   ***  Feature def loader  ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|           Load_XML.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "Load_XML.h"
#include "tinyxml.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *								=== TWeakClassifier ===
 */

/*
 *	TWeakClassifier::TWeakClassifier()
 *		- default constructor
 */
TWeakClassifier::TWeakClassifier()
	:p_alpha(0)
{}

/*
 *	TWeakClassifier::~TWeakClassifier()
 *		- default destructor
 */
TWeakClassifier::~TWeakClassifier()
{
	if(p_alpha)
		delete[] p_alpha;
}

/*
 *	bool TWeakClassifier::operator =(const TWeakClassifier &r_t_classifier)
 *		- copy operator
 *		- returns true on success, false on failure
 */
bool TWeakClassifier::operator =(const TWeakClassifier &r_t_classifier)
{
	if(!p_alpha || n_alpha_num != r_t_classifier.n_alpha_num) {
		if(p_alpha)
			delete[] p_alpha;
		if(!(p_alpha = new float[n_alpha_num = r_t_classifier.n_alpha_num]))
			return false;
	}
	// alloc alphas if necessary

	memcpy(p_alpha, r_t_classifier.p_alpha, n_alpha_num * sizeof(float));
	// copy alphas

	n_position_x = n_position_x;
	n_position_y = n_position_y;
	n_block_a = n_block_a;
	n_block_b = n_block_b;
	n_block_width = n_block_width;
	n_block_height = n_block_height;
	n_width = n_width;
	n_height = n_height;
	// copy other properties

	return true;
}

/*
 *								=== ~TWeakClassifier ===
 */

/*
 *								=== TAdaBoostClassifier ===
 */

/*
 *	TAdaBoostClassifier::TAdaBoostClassifier()
 *		- default constructor
 */
TAdaBoostClassifier::TAdaBoostClassifier()
	:p_s_description(0), p_weak_classifier(0)
{}

/*
 *	TAdaBoostClassifier::~TAdaBoostClassifier()
 *		- default destructor
 */
TAdaBoostClassifier::~TAdaBoostClassifier()
{
	if(p_s_description)
		delete[] p_s_description;
	if(p_weak_classifier)
		delete[] p_weak_classifier;
}

/*
 *	bool TAdaBoostClassifier::Copy(const TAdaBoostClassifier &r_t_classifier)
 *		- copies contents of r_t_classifier to this
 *		- returns true on success, false on failure
 */
bool TAdaBoostClassifier::Copy(const TAdaBoostClassifier &r_t_classifier)
{
	if(p_s_description)
		delete[] p_s_description;
	if(p_weak_classifier) {
		delete[] p_weak_classifier;
		p_weak_classifier = 0;
	}
	// delete original contents if any

	if(r_t_classifier.p_s_description) {
		if(!(p_s_description = new char[strlen(r_t_classifier.p_s_description) + 1]))
			return false;
		strcpy(p_s_description, r_t_classifier.p_s_description);
	} else
		p_s_description = 0;
	// copy description

	if(r_t_classifier.p_weak_classifier) {
		n_weak_classifier_num = r_t_classifier.n_weak_classifier_num;
		if(!(p_weak_classifier = new TWeakClassifier[n_weak_classifier_num]))
			return false;
		for(int i = 0; i < n_weak_classifier_num; ++ i) {
			if(!(p_weak_classifier[i] = r_t_classifier.p_weak_classifier[i]))
				return false;
		}
	} else {
		n_weak_classifier_num = 0;
		p_weak_classifier = 0;
	}
	// copy weak classifiers

	f_final_thresh = r_t_classifier.f_final_thresh;
	f_cutoff_thresh = r_t_classifier.f_cutoff_thresh;
	// copy other properties

	return true;
}

/*
 *	bool TAdaBoostClassifier::Load_XML(const char *p_s_filename)
 *		- loads classifier from XML (Michal Hradis's format)
 *		- returns true on success, false on failure
 *		- note original contents of this classifier are deleted
 */
bool TAdaBoostClassifier::Load_XML(const char *p_s_filename)
{
	__FuncGuard("TAdaBoostClassifier::Load_XML");

	if(p_s_description) {
		delete[] p_s_description;
		p_s_description = 0;
	}
	if(p_weak_classifier) {
		delete[] p_weak_classifier;
		p_weak_classifier = 0;
	}
	// delete original contents if any

	TiXmlDocument document(p_s_filename);
	if(!document.LoadFile())
		return false;
	// parse XML

	const TiXmlNode *p_training;
	if(!(p_training = document.FirstChild("trainingAdaBoost")))
		return false;

	const TiXmlElement *p_classifier;
	if(!(p_classifier = p_training->FirstChildElement("AdaBoostClassifier")))
		return false;

	const char *p_s_name = p_classifier->Attribute("description");
	// get description

	if(p_s_name) {
		if(!(p_s_description = new char[strlen(p_s_name) + 1]))
			return false;
		strcpy(p_s_description, p_s_name);
	} else
		p_s_description = 0;
	// copy description

	int n_classifier_num = 0;
	int n_min_index = (unsigned int)-1 / 2;
	int n_max_index = -1;

	for(const TiXmlNode *p_weak_cl_node = p_classifier->FirstChild("WeakClassifier");
	   p_weak_cl_node; p_weak_cl_node = p_classifier->IterateChildren(p_weak_cl_node)) {
		const TiXmlElement *p_weak_cl;
		if(!(p_weak_cl = p_weak_cl_node->ToElement()))
			continue;

		const char *p_s_index;
		if(!(p_s_index = p_weak_cl->Attribute("index")))
			return false;
		int n_index = atol(p_s_index);
		if(n_min_index > n_index)
			n_min_index = n_index;
		if(n_max_index < n_index)
			n_max_index = n_index;
		++ n_classifier_num;
		// count classifiers
	}
	// iterate trough weak classifiers

	if(n_min_index != 0 || n_max_index != n_classifier_num - 1)
		return false;
	// index mismatch

	f_final_thresh = 0;
	f_cutoff_thresh = -1e10;
	// not saved for now

	n_weak_classifier_num = n_classifier_num;
	if(!(p_weak_classifier = new TWeakClassifier[n_classifier_num]))
		return false;
	// alloc classifiers

	for(const TiXmlNode *p_weak_cl_node = p_classifier->FirstChild("WeakClassifier");
	   p_weak_cl_node; p_weak_cl_node = p_classifier->IterateChildren(p_weak_cl_node)) {
		const TiXmlElement *p_weak_cl;
		if(!(p_weak_cl = p_weak_cl_node->ToElement()))
			continue;

		const char *p_s_index;
		if(!(p_s_index = p_weak_cl->Attribute("index")))
			return false;
		int n_index = atol(p_s_index);
		_ASSERTE(n_index >= 0 && n_index < n_classifier_num);
		// get classifier index

		TWeakClassifier &r_t_classifier = p_weak_classifier[n_index];
		// destination classifier

		const TiXmlElement *p_feature;
		if(!(p_feature = p_weak_cl->FirstChildElement("DiscreteFeature")))
			return false;

		const char *p_s_index_range;
		if(!(p_s_index_range = p_feature->Attribute("indexRange")))
			return false;

		r_t_classifier.n_alpha_num = atol(p_s_index_range);
		if(!(r_t_classifier.p_alpha = new float[r_t_classifier.n_alpha_num]))
			return false;
		// alloc alphas

		for(int i = 0; i < r_t_classifier.n_alpha_num; ++ i) {
			char p_s_alpha_name[32];
			sprintf(p_s_alpha_name, "alpha%d", i);

			const char *p_s_alpha;
			if(!(p_s_alpha = p_feature->Attribute(p_s_alpha_name)))
				return false;
			r_t_classifier.p_alpha[i] = atof(p_s_alpha);
		}
		// read alphas

		const TiXmlElement *p_ab_distance;
		if(!(p_ab_distance = p_feature->FirstChildElement("TFeatureABDistance")))
			return false;

		const char *p_s_tmp;
		if(!(p_s_tmp = p_ab_distance->Attribute("positionX")))
			return false;
		r_t_classifier.n_position_x = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("positionY")))
			return false;
		r_t_classifier.n_position_y = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blockA")))
			return false;
		r_t_classifier.n_block_a = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blockB")))
			return false;
		r_t_classifier.n_block_b = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blockWidth")))
			return false;
		r_t_classifier.n_block_width = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blockHeight")))
			return false;
		r_t_classifier.n_block_height = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blocksX")))
			return false;
		r_t_classifier.n_width = atol(p_s_tmp);
		if(!(p_s_tmp = p_ab_distance->Attribute("blocksY")))
			return false;
		r_t_classifier.n_height = atol(p_s_tmp);
		// read weak classifier properties
	}
	// iterate trough weak classifiers

	return true;
}

/*
 *								=== ~TAdaBoostClassifier ===
 */
