/*
								+----------------------------------+
								|                                  |
								|  ***  GPU ada-boost engine  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <math.h>
#include <algorithm>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/VertexBufferObject.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "../../berLame_src/TextureUtil.h"
#include "../../berLame_src/Vector.h"
#include "../../berLame_src/Timer.h"
#include "Load_XML.h"
#include "ConvPlan.h"
#include "RasCalib.h"
#include "Classifier.h"

static Vector2f v_mouse(0, 0);
// mouse position in OpenGL coords

static int n_width = 512, n_height = 512;
// window size

static CGLState *p_state = 0;
// OpenGL state guard

static CGLTexture_2D *p_src_texture = 0;

static TConvolutionList t_list;
static CConvolutionManager man;
// convolution manager

static int n_show_image = -1, n_slice = 0;

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera
}

void onDisplay()
{
	p_state->ClearColor4f(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	p_state->DisableTexture2D();
	p_state->DisableTexture3D();

	float f_size_x;
	float f_size_y;
	float f_r = 0;
	if(n_show_image < 0) {
		p_src_texture->Bind_Enable(p_state);
		f_size_x = 1.0f / n_width * p_src_texture->n_Width();
		f_size_y = 1.0f / n_height * p_src_texture->n_Height();
	} else {
		const CGLTexture_3D &r_texture = *man.r_Convolution_Image(n_show_image).p_texture;
		r_texture.Bind_Enable(p_state);
		f_size_x = 1.0f / n_width * r_texture.n_Width();
		f_size_y = 1.0f / n_height * r_texture.n_Height();
		f_r = 1.0f / r_texture.n_Depth() * n_slice;
	}

	glBegin(GL_QUADS);
	glTexCoord3f(0, 0, f_r);
	glVertex2f(-f_size_x, f_size_y);

	glTexCoord3f(1, 0, f_r);
	glVertex2f(f_size_x, f_size_y);

	glTexCoord3f(1, 1, f_r);
	glVertex2f(f_size_x, -f_size_y);

	glTexCoord3f(0, 1, f_r);
	glVertex2f(-f_size_x, -f_size_y);
	glEnd();

	glutSwapBuffers();
	// display results
}

static int n_button;

void onMouseClick(int _n_button, int n_state, int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	n_button = _n_button;

	switch(_n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	switch(n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onPassiveMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;
	glutPostRedisplay();
}

void onKeyPressed(unsigned char n_key_code, int n_x, int n_y)
{
	switch(n_key_code) {
	case 27:
		if(p_state)
			delete p_state;
		exit(0);
		break;
	case 'n':
	case 'b':
		if(n_key_code == 'n' && n_show_image + 1 < man.n_Convolution_Num())
			++ n_show_image;
		else if(n_key_code == 'b' && n_show_image > -1)
			-- n_show_image;
		else
			return;
		n_slice = 0;
		printf("showing convolution %d, slice %d", n_show_image, n_slice);
		if(n_show_image >= 0)
			t_list.convolution_list[n_show_image].Dump();
		else
			printf("\t(source image)\n");
		glutPostRedisplay();
		break;
	case 'c':
	case 'v':
		{
			int n_slice_num = t_list.convolution_list[n_show_image].n_block_width *
				t_list.convolution_list[n_show_image].n_block_height;
			if(n_key_code == 'c' && n_slice < n_slice_num - 1)
				++ n_slice;
			else if(n_key_code == 'v' && n_slice > 0)
				-- n_slice;
			else
				return;
			printf("showing convolution %d, slice %d", n_show_image, n_slice);
			if(n_show_image >= 0)
				t_list.convolution_list[n_show_image].Dump();
			else
				printf("\t(source image)\n");
			glutPostRedisplay();
			break;
		}
	default:
		break;
	}
}

CGLTexture_2D *p_My_LoadTexture_Jpeg(CGLState *p_state, const char *p_s_filename,
	bool b_mipmaps = true, bool b_clamp_to_border = false)
{
	TBmp *p_bitmap;

	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb")))
		return 0;
	CJpegDecoder decoder;
	if(!(p_bitmap = decoder.p_Decode_JPEG(p_fr))) {
		fclose(p_fr);
		return 0;
	}
	fclose(p_fr);

	CGLTexture_2D *p_texture;
	if(!(p_texture = new CGLTexture_2D(p_state, p_bitmap->n_width, p_bitmap->n_height,
	   GL_LUMINANCE8, b_mipmaps, 0, GL_BGRA, GL_UNSIGNED_BYTE, p_bitmap->p_buffer))) {
		delete[] p_bitmap->p_buffer;
		delete p_bitmap;
		return 0;
	}
	delete[] p_bitmap->p_buffer;
	delete p_bitmap;

	if(b_clamp_to_border) {
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_BORDER);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_BORDER);
	}

	if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
		delete p_texture;
		return 0;
	}

	return p_texture;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("GPU ada-boost engine");
	// init OpenGL using GLUT

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create OpenGL state guard

	printf("=== Checking OpenGL support ===\n");
	if(!CGLExtensionHandler::n_GetGL12FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(1, 2))
		printf("\tOpenGL 1.2\n");
	else {
		fprintf(stderr, "\terror: OpenGL 1.2 not supported\n");
		return -1; // need 3D texturing
	}
	if(!CGLExtensionHandler::n_GetGL13FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(1, 3))
		printf("\tOpenGL 1.3\n");
	else
		fprintf(stderr, "\twarning: OpenGL 1.3 not supported\n");
	if(!CGLExtensionHandler::n_GetGL14FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(1, 4))
		printf("\tOpenGL 1.4\n");
	else
		fprintf(stderr, "\twarning: OpenGL 1.4 not supported\n");
	if(!CGLExtensionHandler::n_GetGL15FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(1, 5))
		printf("\tOpenGL 1.5\n");
	else
		fprintf(stderr, "\twarning: OpenGL 1.5 not supported\n");
	bool b_have_gl20 = false;
	if(!CGLExtensionHandler::n_GetGL20FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(2, 0)) {
		printf("\tOpenGL 2.0\n");
		b_have_gl20 = CGL_Core_shader::b_Supported();
	} else
		fprintf(stderr, "\twarning: OpenGL 2.0 not supported\n");
	if(!CGLExtensionHandler::n_GetGL21FuncPointers() &&
	   CGLExtensionHandler::b_Support_OpenGL(2, 1))
		printf("\tOpenGL 2.1\n");
	else {
		fprintf(stderr, "\twarning: OpenGL 2.1 not supported"
			" (have to deal with NVidia render to 3D-texture bug)\n");
	}
	if(CGL_ARB_shader::b_Supported())
		printf("\tGL_ARB_fragment_shader\n");
	else if(!b_have_gl20) {
		fprintf(stderr, "\terror: GL_ARB_fragment_shader not supported\n");
		return -1;
	} else
		fprintf(stderr, "\twarning: GL_ARB_fragment_shader not supported\n");
	if(CGLRenderBuffer_FBO::b_Supported())
		printf("\tGL_EXT_framebuffer_object\n");
	else {
		fprintf(stderr, "\terror: GL_EXT_framebuffer_object not supported\n");
		return -1;
	}
	// check necessary OpenGL extensions

	TAdaBoostClassifier t_classifier;
	printf("\n=== Loading XML classifier ===\n");
	if(!(t_classifier.Load_XML("TLRD-11-22-42-24.xml"))) {
		fprintf(stderr, "error: error loading XML\n");
		return -1;
	}
	printf("\tAdaBoost classifier \'%s\'\n", (t_classifier.p_s_description)?
		t_classifier.p_s_description : "(null)");
	printf("\t%d weak classifiers\n\n", t_classifier.n_weak_classifier_num);
	// load classifier from XML

	if(!t_list.Create(t_classifier.n_weak_classifier_num, t_classifier.p_weak_classifier)) {
		fprintf(stderr, "error: error creating convolution list\n");
		return -1;
	}
	t_list.Dump();
	// create list of all necessary convolutions

	printf("\n=== Load test image ===\n");
	if(!(p_src_texture = p_My_LoadTexture_Jpeg(p_state, "conv_test_256.jpg", true, true))) {
		fprintf(stderr, "error: error loading test image\n");
		return -1;
	} else
		printf("\tdone\n");
	// load src image

	printf("\n=== Create convolution plan ===\n");
	const bool b_16_bit_precision = true;
	printf("\tconvolution precision is %d bits\n", (b_16_bit_precision)? 16 : 8);
#ifdef RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM
	printf("\twarning: rasterizer calibration using GL_ARB_fragment_program\n");
#else
	printf("\trasterizer calibration using GL_ARB_fragment_shader\n");
#endif
	if(!man.CreateConvolutionPlan(p_state, t_list, p_src_texture->n_Width(),
	   p_src_texture->n_Height(), b_16_bit_precision)) {
		fprintf(stderr, "error: error creating convolution plan\n");
		return -1;
	} else
		printf("\tdone\n");
	// create convolution plan

	printf("\n=== Validate convolution ===\n");
	{
		int n_max;
		float f_dev;
		if(!man.Validate(p_state, p_src_texture, n_max, f_dev))
			fprintf(stderr, "\terror: didn't pass validation\n");
		else {
			printf("\tdone; convolutions of test image come out precise enough (maximal\n"
				"\tdifference is %d, standard deviation is %g; in 0 - 255 scale)\n", n_max, f_dev);
		}
	}
	// see how precise the results are

	printf("\n=== Create & validate AdaBoost plan ===\n");

	CGPU_AdaBoost ab(p_state, p_src_texture->n_Width(),
		p_src_texture->n_Height(), t_classifier, man);

	{
		int n_max;
		float f_dev;
		if(!ab.Validate_ClassifierSampling(p_state, p_src_texture, man, n_max, f_dev))
			fprintf(stderr, "\terror: didn't pass validation\n");
		else {
			printf("\tdone; classifier sampling of test image come out precise enough (maximal\n"
				"\tdifference is %d, standard deviation is %g; in 0 - 255 scale)\n", n_max, f_dev);
		}
	}

	printf("\n=== Calculate test image convolutions ===\n");
	CTimer timer;
	timer.ResetTimer();
	const int n_pass_num = 400;
	for(int i = 0; i < n_pass_num; ++ i) {
		if(!man.Convolve(p_state, p_src_texture)) {
			fprintf(stderr, "error: error executing convolution plan\n");
			break;
		} else if(i + 1 == n_pass_num)
			printf("\tdone\n");
	}
	glFlush();
	glFinish(); // make sure no frames are left in queue
	float f_time = timer.f_Time();
	printf("\trunning at %g frames/sec (%g frame time)\n",
		n_pass_num / f_time, f_time / n_pass_num);
	// run a little benchmark

	if(glGetError() != GL_NO_ERROR)
		fprintf(stderr, "\terror: OpenGL erorr(s) occured\n");
	else
		printf("\tok ... no OpenGL erorrs occured\n");

	printf("\nnow you can switch between original image and different kernels using\n"
		"\'b\' and \'n\' or between different shifts of kernel using \'c\' and \'v\'\n\n");

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	glutKeyboardFunc(onKeyPressed);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
