/*
								+-----------------------------------+
								|                                   |
								|  ***  Rasterizer calibrator  ***  |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|            RasCalib.h             |
								|                                   |
								+-----------------------------------+
*/

#ifndef __RASTERIZER_CALIBRATOR_INCLUDED
#define __RASTERIZER_CALIBRATOR_INCLUDED

/*
 *	RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM
 *		- if defined, fragment program is used instead of fragment shader
 *		- note it might theoretically cause problems in case fragment shader
 *		  was getting different texcoords than fragment program and vice versa
 *		  so it's advisable to use the same shadiing language to calibrate
 *		  rasterizer and later to process textures
 */
#define RASTERIZER_CALIBRATOR_USE_FRAGMENT_PROGRAM

class CRasterizerCalibrator {
protected:
	static TShaderInfo display_st_shader;

public:
	/*
	 *	struct TTexCoordRect
	 *		- texture coordinates for axis aligned rectangle
	 */
	struct TTexCoordRect {
		float f_min_s, f_min_t;
		float f_max_s, f_max_t;

		TTexCoordRect()
		{}

		TTexCoordRect(float f_min, float f_max)
			:f_min_s(f_min), f_min_t(f_min),
			f_max_s(f_max), f_max_t(f_max)
		{}

		TTexCoordRect(float _f_min_s, float _f_min_t, float _f_max_s, float _f_max_t)
			:f_min_s(_f_min_s), f_min_t(_f_min_t),
			f_max_s(_f_max_s), f_max_t(_f_max_t)
		{}
	};

	/*
	 *	static bool CRasterizerCalibrator::CalibrateRasterizer(CGLState *p_state,
	 *		int n_width, int n_height, TTexCoordRect t_in_texcoords,
	 *		TTexCoordRect &r_t_out_texcoords, float f_epsilon, int n_max_steps = 100)
	 *		- determines correct texcoords that has to be passed to OpenGL so
	 *		  n_width per n_height pixels quad has desired texcoords in it's corner pixels
	 *		- p_state is OpenGL state guard
	 *		- r_t_in_texcoords are desired texcoords, r_t_out_texcoords are correct texcoords
	 *		- f_epsilon is maximal texcoord deviation
	 *		- the method iteratively converges to solution, n_max_steps is maximal number of
	 *		  render passes before it either finds solution or fails
	 *		- returns true on success, false on failure
	 */
	static bool CalibrateRasterizer(CGLState *p_state, int n_width, int n_height,
		TTexCoordRect t_in_texcoords, TTexCoordRect &r_t_out_texcoords,
		float f_epsilon, int n_max_steps = 100);

protected:
	/*
	 *	static float CRasterizerCalibrator::f_fabs(float f_x)
	 *		- returns absolute value of f_x
	 *		- want to avoid conversion to double and back again
	 *		- note this is most likely not going to compile to effective code on GCC
	 *
	 *		- not used anymore
	 */
	static float f_fabs(float f_x);

	/*
	 *	static inline bool CRasterizerCalibrator::b_zero(float f_x, float f_epsilon)
	 *		- returns true in case f_x is zero, with tolerance f_epsilon;
	 *		  otherwise returns false
	 */
	static inline bool b_zero(float f_x, float f_epsilon);
};

#endif //__RASTERIZER_CALIBRATOR_INCLUDED
