/*
								+-----------------------------------+
								|                                   |
								| ***  Heap object scope guard  *** |
								|                                   |
								|   Copyright   -tHE SWINe- 2007   |
								|                                   |
								|           ScopeGuard.h            |
								|                                   |
								+-----------------------------------+
*/

#ifndef __SCOPE_GUARD_INCLUDED
#define __SCOPE_GUARD_INCLUDED

/*
 *	note this has tedious syntax CScalarScopeGuard<type*> sg(object);
 *	but it helps anyway ... maybe one day i'm going to write some kind
 *	of adaptor function for that
 */

/*
 *	template <class CPointerType>
 *	class CScopeGuard
 *		- base template class for simple automatic cleanup of objects on heap
 */
template <class CPointerType>
class CScopeGuard {
protected:
	CPointerType m_p_pointer;

public:
	/*
	 *	inline CScopeGuard::CScopeGuard(CPointerType p_pointer)
	 *		- default constructor
	 *		- specifies pointer of object to be deleted when guard reaches
	 *		  the end of it's visibility scope
	 */
	inline CScopeGuard(CPointerType p_pointer)
		:m_p_pointer(p_pointer)
	{}

	/*
	 *	void CScopeGuard::Disable()
	 *		- disables scope guard so the object specified in constructor
	 *		  is not going to be deleted on the end of the guard's scope
	 */
	void Disable()
	{
		m_p_pointer = 0;
	}
};

/*
 *	template <class CPointerType>
 *	class CScalarScopeGuard
 *		- template class for simple automatic cleanup of scalar objects on heap
 */
template <class CPointerType>
class CScalarScopeGuard : public CScopeGuard<CPointerType> {
public:
	/*
	 *	inline CScalarScopeGuard::CScalarScopeGuard(CPointerType p_pointer)
	 *		- default constructor
	 *		- specifies pointer of object to be deleted when the guard reaches
	 *		  the end of it's visibility scope
	 */
	inline CScalarScopeGuard(CPointerType p_pointer)
		:CScopeGuard<CPointerType>(p_pointer)
	{}

	/*
	 *	CScalarScopeGuard::CScalarScopeGuard(CScalarScopeGuard &r_deleter)
	 *		- copy-constructor
	 */
	inline CScalarScopeGuard(CScalarScopeGuard &r_deleter)
		:m_p_pointer(r_deleter.m_p_pointer)
	{
		r_deleter.Disable();
	}

	/*
	 *	inline CScalarScopeGuard::~CScalarScopeGuard()
	 *		- destructor; deletes object specified in constructor, unless Disable() was called
	 */
	inline ~CScalarScopeGuard()
	{
		if(m_p_pointer)
			delete m_p_pointer;
	}
};

/*
 *	template <class CPointerType>
 *	class CVectorScopeGuard
 *		- template class for simple automatic cleanup of object arrays on heap
 */
template <class CPointerType>
class CVectorScopeGuard : public CScopeGuard<CPointerType> {
public:
	/*
	 *	inline CVectorScopeGuard::CVectorScopeGuard(CPointerType p_pointer)
	 *		- default constructor
	 *		- specifies pointer to array of objects to be deleted when
	 *		  the guard reaches the end of it's visibility scope
	 */
	inline CVectorScopeGuard(CPointerType p_pointer)
		:CScopeGuard<CPointerType>(p_pointer)
	{}

	/*
	 *	CVectorScopeGuard::CVectorScopeGuard(CVectorScopeGuard &r_deleter)
	 *		- copy-constructor
	 */
	inline CVectorScopeGuard(CVectorScopeGuard &r_deleter)
		:m_p_pointer(r_deleter.m_p_pointer)
	{
		r_deleter.Disable();
	}

	/*
	 *	inline CVectorScopeGuard::~CVectorScopeGuard()
	 *		- destructor; deletes object specified in constructor, unless Disable() was called
	 */
	inline ~CVectorScopeGuard()
	{
		if(m_p_pointer)
			delete[] m_p_pointer;
	}
};

#endif //__SCOPE_GUARD_INCLUDED
